unit sfSDist;

{Common code for special functions: Statistical distributions}

interface

{$i std.inc}

{$ifdef BIT16}
{$N+}
{$endif}

{$ifdef NOBASM}
  {$undef BASM}
{$endif}


(*************************************************************************

 DESCRIPTION   :  Common code for special functions: Statistical distributions

 REQUIREMENTS  :  TP5.5-7, D2-D7/D9-D10/D12/D17-D18, FPC, VP, WDOSX

 EXTERNAL DATA :  ---

 MEMORY USAGE  :  ---

 DISPLAY MODE  :  ---

 REMARK        :  The unit can be compiled with TP5 but some functions generate
                  invalid operations due to TP5's brain-damaged usage of the FPU

 REFERENCES    :  References used in this unit, main index in sfBasic

                  [7] Cephes Mathematical Library, Version 2.8
                      http://www.moshier.net/#Cephes or http://www.netlib.org/cephes/
                 [19] Boost C++ Libraries, Release 1.42.0, 2010.
                      http://www.boost.org/
                 [21] GNU Scientific Library, GSL-1.14 (March 2010),
                      http://www.gnu.org/software/gsl
                 [39] R: A Language and Environment for Statistical Computing,
                      Version 2.11.1, http://www.r-project.org/

                  Not used for actual implementation but for background info:

                      H. Rinne, Taschenbuch der Statistik, 4.Auflage, Frankfurt 2008

                      Wikipedia, the free encyclopedia. List of probability distributions
                      with links to many specific probability distributions:
                      http://en.wikipedia.org/wiki/List_of_probability_distributions


 Version  Date      Author      Modification
 -------  --------  -------     ------------------------------------------
 1.00.00  17.08.10  W.Ehrhardt  Common version number after split
 1.00.01  05.09.10  we          sfc_chi2_pdf/cdf/inv
 1.00.02  07.09.10  we          sfc_gamma_pdf/cdf/inv
 1.00.03  08.09.10  we          Improved arg checking and NAN/INF handling

 1.05.00  13.04.11  we          sfc_cauchy_pdf/cdf/inv
 1.05.01  14.04.11  we          sfc_normal_pdf/cdf/inv
 1.05.02  14.04.11  we          sfc_exp_pdf/cdf/inv
 1.05.03  15.04.11  we          sfc_lognormal_pdf/cdf/inv
 1.05.04  16.04.11  we          sfc_logistic_pdf/cdf/inv
 1.05.05  16.04.11  we          sfc_weibull_pdf/cdf/inv
 1.05.05  17.04.11  we          sfc_laplace_pdf/cdf/inv
 1.05.06  17.04.11  we          changed sh,sc to a,b in sfc_gamma_pdf/cdf/inv
 1.05.07  19.04.11  we          sfc_pareto_pdf/cdf/inv
 1.05.08  20.04.11  we          sfc_uniform_pdf/cdf/inv

 1.08.00  15.08.11  we          Fix sfc_t_cdf for t>0 and large nu
 1.08.01  15.08.11  we          sfc_normstd_inv uses sfc_erfc_inv

 1.10.00  10.12.11  we          sfc_triangular_pdf/cdf/inv
 1.10.01  13.12.11  we          sfc_beta_pdf uses sfc_ibetaprefix
 1.10.02  14.12.11  we          sfc_binomial_cdf/pmf
 1.10.03  15.12.11  we          sfc_poisson_cdf/pmf
 1.10.04  18.12.11  we          sfc_negbinom_cdf/pmf
 1.10.05  26.12.11  we          sfc_hypergeo_cdf/pmf
 1.10.06  27.12.11  we          Fix sfc_hypergeo_pmf(0,0,0,k)

 1.17.00  07.04.13  we          Improved sfc_chi2_pdf for very small x
 1.17.01  13.04.13  we          k: longint in sfc_poisson_pmf
 1.17.02  14.04.13  we          sfc_rayleigh_pdf/cdf/inv
 1.17.03  19.04.13  we          sfc_maxwell_pdf/cdf/inv
 1.17.04  20.04.13  we          sfc_evt1_pdf/cdf/inv
 1.17.05  20.04.13  we          sfc_maxwell_cdf/inv with lower igamma functions

 1.20.00  14.08.13  we          Removed code fragment for y>1 in sfc_beta_inv
 1.20.01  14.08.13  we          Check a,b > 0 in sfc_beta_cdf
 1.20.02  15.08.13  we          sfc_kumaraswamy_pdf/cdf/inv
 1.20.03  17.08.13  we          Improved sfc_lognormal_pdf with expmx2h
 1.20.04  18.08.13  we          normal/normstd_pdf/cdf with erf_z and erf_p

***************************************************************************)


(*-------------------------------------------------------------------------
 (C) Copyright 2009-2013 Wolfgang Ehrhardt

 This software is provided 'as-is', without any express or implied warranty.
 In no event will the authors be held liable for any damages arising from
 the use of this software.

 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:

 1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software in
    a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

 2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

 3. This notice may not be removed or altered from any source distribution.
----------------------------------------------------------------------------*)

(*-------------------------------------------------------------------------
  This Pascal code uses material and ideas from open source and public
  domain libraries, see the file '3rdparty.ama' for the licenses.
---------------------------------------------------------------------------*)


function sfc_beta_pdf(a, b, x: extended): extended;
  {-Return the probability density function of the beta distribution with}
  { parameters a and b: sfc_beta_pdf = x^(a-1)*(1-x)^(b-1) / beta(a,b)}

function sfc_beta_cdf(a, b, x: extended): extended;
  {-Return the cumulative beta distribution function, a>0, b>0}

function sfc_beta_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the beta distribution function. a>0, b>0;}
  { 0 <= y <= 1. Given y the function finds x such that beta_cdf(a, b, x) = y}

function sfc_binomial_cdf(p: extended; n, k: longint): extended;
  {-Return the cumulative binomial distribution function with number}
  { of trials n >= 0 and success probability 0 <= p <= 1}

function sfc_binomial_pmf(p: extended; n, k: longint): extended;
  {-Return the binomial distribution probability mass function with number}
  { of trials n >= 0 and success probability 0 <= p <= 1}

function sfc_cauchy_pdf(a, b, x: extended): extended;
  {-Return the Cauchy probability density function with location a }
  { and scale b > 0, 1/(Pi*b*(1+((x-a)/b)^2))}

function sfc_cauchy_cdf(a, b, x: extended): extended;
  {-Return the cumulative Cauchy distribution function with location a}
  { and scale b > 0, = 1/2 + arctan((x-a)/b)/Pi}

function sfc_cauchy_inv(a, b, y: extended): extended;
  {-Return the functional inverse of Cauchy distribution function}
  { with location a and scale b > 0}

function sfc_chi2_pdf(nu: longint; x: extended): extended;
  {-Return the probability density function of the chi-square distribution, nu>0, x >= 0}

function sfc_chi2_cdf(nu: longint; x: extended): extended;
  {-Return the cumulative chi-square distribution with nu>0 degrees of freedom, x >= 0}

function sfc_chi2_inv(nu: longint; p: extended): extended;
  {-Return the functional inverse of the chi-square distribution, nu>0, 0 <= p < 1}

function sfc_evt1_pdf(a, b, x: extended): extended;
  {-Return the probability density function of the Extreme Value Type I distribution}
  { with location a and scale b > 0, result = exp(-(x-a)/b)/b * exp(-exp(-(x-a)/b)) }

function sfc_evt1_cdf(a, b, x: extended): extended;
  {-Return the cumulative Extreme Value Type I distribution function}
  { with location a and scale b > 0;  result = exp(-exp(-(x-a)/b)). }

function sfc_evt1_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the Extreme Value Type I distribution}
  { function with location a and scale b > 0;  result = a - b*ln(ln(-y)). }

function sfc_exp_pdf(a, alpha, x: extended): extended;
  {-Return the exponential probability density function with location a }
  { and rate alpha > 0, = alpha*exp(-alpha*(x-a)) if x >= a, 0 if x < a.}

function sfc_exp_cdf(a, alpha, x: extended): extended;
  {-Return the cumulative exponential distribution function with location a}
  { and rate alpha > 0, = 1 - exp(-alpha*(x-a)) if x >= a, 0 if x < a.}

function sfc_exp_inv(a, alpha, y: extended): extended;
  {-Return the functional inverse of exponential distribution function with}
  { location a and rate alpha > 0}

function sfc_f_pdf(nu1, nu2: longint; x: extended): extended;
  {-Return the probability density function of the F distribution; x >= 0, nu1, nu2 > 0}

function sfc_f_cdf(nu1, nu2: longint; x: extended): extended;
  {-Return the cumulative F distribution function; x >= 0, nu1, nu2 > 0}

function sfc_f_inv(nu1, nu2: longint; y: extended): extended;
  {-Return the functional inverse of the F distribution, nu1, nu2 > 0, 0 <= y <= 1}

function sfc_gamma_pdf(a, b, x: extended): extended;
  {-Return the probability density function of a gamma distribution with shape}
  { a>0, scale b>0: sfc_gamma_pdf = x^(a-1)*exp(-x/b)/gamma(a)/b^a, x>0}

function sfc_gamma_cdf(a, b, x: extended): extended;
  {-Return the cumulative gamma distribution function, shape a>0, scale b>0}

function sfc_gamma_inv(a, b, p: extended): extended;
  {-Return the functional inverse of the gamma distribution function, shape a>0,}
  { scale b>0, 0 <= p <= 1, i.e. finds x such that gamma_cdf(a, b, x) = p}

function sfc_hypergeo_pmf(n1,n2,n,k: longint): extended;
  {-Return the hypergeometric distribution probability mass function; n,n1,n2 >= 0, n <= n1+n2;}
  { i.e. the probability that among n randomly chosen samples from a container}
  { with n1 type1 objects and n2 type2 objects are exactly k type1 objects:}

function sfc_hypergeo_cdf(n1,n2,n,k: longint): extended;
  {-Return the cumulative hypergeometric distribution function; n,n1,n2 >= 0, n <= n1+n2}

function sfc_kumaraswamy_pdf(a, b, x: extended): extended;
  {-Return the Kumaraswamy probability density function with shape}
  { parameters a,b>0, 0<=x<=1; result = a*b*x^(a-1)*(1-x^a)^(b-1) }

function sfc_kumaraswamy_cdf(a, b, x: extended): extended;
  {-Return the cumulative Kumaraswamy distribution function with}
  { shape parameters a,b > 0, 0 <= x <= 1; result = 1-(1-x^a)^b}

function sfc_kumaraswamy_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the Kumaraswamy distribution}
  { with shape parameters a,b > 0; result = [1-(1-y)^(1/b)]^(1/a)}

function sfc_laplace_pdf(a, b, x: extended): extended;
  {-Return the Laplace probability density function with location a}
  { and scale b > 0, result = exp(-abs(x-a)/b) / (2*b)}

function sfc_laplace_cdf(a, b, x: extended): extended;
  {-Return the cumulative Laplace distribution function with location a and scale b > 0}

function sfc_laplace_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the Laplace distribution with location a and scale b > 0}

function sfc_logistic_pdf(a, b, x: extended): extended;
  {-Return the logistic probability density function with location a}
  { and scale parameter b > 0, exp(-(x-a)/b)/b/(1+exp(-(x-a)/b))^2}

function sfc_logistic_cdf(a, b, x: extended): extended;
  {-Return the cumulative logistic distribution function with}
  { location a and scale parameter b > 0}

function sfc_logistic_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the logistic distribution}
  { with location a and scale parameter b > 0}

function sfc_lognormal_pdf(a, b, x: extended): extended;
  {-Return the log-normal probability density function with}
  { location a and scale parameter b > 0, zero for x <= 0.}

function sfc_lognormal_cdf(a, b, x: extended): extended;
  {-Return the cumulative log-normal distribution function with}
  { location a and scale parameter b > 0, zero for x <= 0,}

function sfc_lognormal_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the log-normal distribution}
  { with location a and scale parameter b > 0, 0 < y < 1.}

function sfc_maxwell_pdf(b, x: extended): extended;
  {-Return the Maxwell probability density function with scale b > 0, x >= 0}

function sfc_maxwell_cdf(b, x: extended): extended;
  {-Return the cumulative Maxwell distribution function with scale b > 0, x >= 0}

function sfc_maxwell_inv(b, y: extended): extended;
  {-Return the functional inverse of the Maxwell distribution with scale b > 0}

function sfc_negbinom_cdf(p,r: extended; k: longint): extended;
  {-Return the cumulative negative binomial distribution function with target}
  { for number of successful trials r > 0 and success probability 0 <= p <= 1}

function sfc_negbinom_pmf(p,r: extended; k: longint): extended;
  {-Return the negative binomial distribution probability mass function with target}
  { for number of successful trials r > 0 and success probability 0 <= p <= 1}

function sfc_normal_pdf(mu, sd, x: extended): extended;
  {-Return the normal (Gaussian) probability density function with mean mu}
  { and standard deviation sd>0, exp(-0.5*(x-mu)^2/sd^2) / sqrt(2*Pi*sd^2)}

function sfc_normal_cdf(mu, sd, x: extended): extended;
  {-Return the normal (Gaussian) distribution density function}
  { with mean mu and standard deviation sd > 0}

function sfc_normal_inv(mu, sd, y: extended): extended;
  {-Return the functional inverse of normal (Gaussian) distribution}
  { with mean mu and standard deviation sd > 0, 0 < y < 1.}

function sfc_normstd_pdf(x: extended): extended;
  {-Return the std. normal probability density function exp(-x^2/2)/sqrt(2*Pi)}

function sfc_normstd_cdf(x: extended): extended;
  {-Return the standard normal distribution function}

function sfc_normstd_inv(y: extended): extended;
  {-Return the inverse standard normal distribution function, 0 < y < 1.}
  { For x=normstd_inv(y) and y from (0,1), normstd_cdf(x) = y}

function sfc_poisson_cdf(mu: extended; k: longint): extended;
  {-Return the cumulative Poisson distribution function with mean mu >= 0}

function sfc_poisson_pmf(mu: extended; k: longint): extended;
  {-Return the Poisson distribution probability mass function with mean mu >= 0}

function sfc_pareto_pdf(k, a, x: extended): extended;
  {-Return the Pareto probability density function with minimum value k > 0}
  { and shape a, x >= a > 0, result = (a/x)*(k/x)^a}

function sfc_pareto_cdf(k, a, x: extended): extended;
  {-Return the cumulative Pareto distribution function minimum value k > 0}
  { and shape a, x >= a > 0, result = 1-(k/x)^a}

function sfc_pareto_inv(k, a, y: extended): extended;
  {-Return the functional inverse of the Pareto distribution with minimum}
  { value k > 0 and shape a, x >= a > 0, result = k/(1-x)^(1/a)}

function sfc_rayleigh_pdf(b, x: extended): extended;
  {-Return the Rayleigh probability density function with}
  { scale b > 0, x >= 0; result = x*exp(-0.5*(x/b)^2)/b^2}

function sfc_rayleigh_cdf(b, x: extended): extended;
  {-Return the cumulative Rayleigh distribution function with scale b > 0, x >= 0}

function sfc_rayleigh_inv(b, y: extended): extended;
  {-Return the functional inverse of the Rayleigh distribution with scale b > 0}

function sfc_t_pdf(nu: longint; x: extended): extended;
  {-Return the probability density function of Student's t distribution, nu>0}

function sfc_t_cdf(nu: longint; t: extended): extended;
  {-Return the cumulative Student t distribution with nu>0 degrees of freedom}

function sfc_t_inv(nu: longint; p: extended): extended;
  {-Return the functional inverse of Student's t distribution, nu>0, 0 <= p <= 1}

function sfc_triangular_pdf(a, b, c, x: extended): extended;
  {-Return the triangular probability density function with}
  { lower limit a, upper limit b, mode c;  a<b, a <= c <= b}

function sfc_triangular_cdf(a, b, c, x: extended): extended;
  {-Return the cumulative triangular distribution function with}
  { lower limit a, upper limit b, mode c;  a<b, a <= c <= b}

function sfc_triangular_inv(a, b, c, y: extended): extended;
  {-Return the functional inverse of the triangular distribution with}
  { lower limit a, upper limit b, mode c; a<b, a <= c <= b, 0 <= y <= 1}

function sfc_uniform_pdf(a, b, x: extended): extended;
  {-Return the uniform probability density function on [a,b], a<b}

function sfc_uniform_cdf(a, b, x: extended): extended;
  {-Return the cumulative uniform distribution function on [a,b], a<b}

function sfc_uniform_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the uniform distribution on [a,b], a<b}

function sfc_weibull_pdf(a, b, x: extended): extended;
  {-Return the Weibull probability density function with shape a > 0}
  { and scale b > 0, result = a*x^(a-1)*exp(-(x/b)^a)/ b^a, x > 0}

function sfc_weibull_cdf(a, b, x: extended): extended;
  {-Return the cumulative Weibull distribution function with}
  { shape parameter a > 0 and scale parameter b > 0}

function sfc_weibull_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the Weibull distribution}
  { shape parameter a > 0 and scale parameter b > 0}


implementation

uses
  AMath,
  sfBasic,
  sfErf,
  sfGamma;


{---------------------------------------------------------------------------}
function sfc_beta_pdf(a, b, x: extended): extended;
  {-Return the probability density function of the beta distribution with}
  { parameters a and b: sfc_beta_pdf = x^(a-1)*(1-x)^(b-1) / beta(a,b)}
var
  p,t: extended;
begin
  if (a <= 0.0) or (b <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_beta_pdf := NaN_x;
    exit;
  end;
  if (x<0.0) or (x>1.0) then sfc_beta_pdf := 0.0
  else if (x=0.0) or (x=1.0) then begin
    p := power(x, a-1.0);
    p := p*power(1.0-x, b-1.0);
    sfc_beta_pdf := p/sfc_beta(a,b);
  end
  else if b=1.0 then sfc_beta_pdf := a*power(x,a-1.0)
  else begin
    {0 < x < 1}
    t := sfc_ibetaprefix(a,b,x,1-x);
    if t<>0 then begin
      p := x*(1.0-x);
      if p*MaxExtended < t then t := PosInf_x else t := t/p;
    end;
    sfc_beta_pdf := t;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_beta_cdf(a, b, x: extended): extended;
  {-Return the cumulative beta distribution function, a>0, b>0}
begin
  if (a <= 0.0) or (b <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_beta_cdf := NaN_x;
    exit;
  end;
  if x<=0.0 then sfc_beta_cdf := 0.0
  else if x>=1.0 then sfc_beta_cdf := 1.0
  else sfc_beta_cdf := sfc_ibeta(a,b,x);
end;


{---------------------------------------------------------------------------}
function sfc_beta_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the beta distribution function. a>0, b>0;}
  { 0 <= y <= 1. Given y the function finds x such that beta_cdf(a, b, x) = y}
var
  aa, bb, y0, yy, d, x, x0, x1, lgm, yp, di, dithresh, yl, yh, xt: extended;
  i, rflg, dir, nflg: integer;

label
  ihalve, newt, breaknewt, done;

begin

  {Ref: Cephes[7], function incbil in ldouble/incbil.c}

  {The routine performs up to 15 Newton iterations to find the root of}
  {ibeta(a,b,x) - y = 0. If Newton failed or is not applicable interval}
  {halving is used. The Pascal input variables are renamed to a,b,y.}

  if (a <= 0.0) or (b <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_beta_inv := NaN_x;
    exit;
  end;

  if (y=0.0) or (y=1.0) then begin
    sfc_beta_inv := y;
    exit;
  end;

  {Copy Pascal input to C variables}
  aa := a;
  bb := b;
  yy := y;

  x0 := 0.0;
  yl := 0.0;
  x1 := 1.0;
  yh := 1.0;
  nflg := 0;

  if (aa <= 1.0) or (bb <= 1.0) then begin
    dithresh := 1e-7;
    rflg := 0;
    y0 := yy;
    x  := a/(a+b);
    y  := sfc_ibeta(a, b, x);
    goto ihalve;
  end
  else dithresh := 1e-4;

  {approximation to inverse function}
  yp := -sfc_normstd_inv(yy);

  if yy > 0.5 then begin
    rflg := 1;
    a  := bb;
    b  := aa;
    y0 := 1.0 - yy;
    yp := -yp;
  end
  else
  begin
    rflg := 0;
    a  := aa;
    b  := bb;
    y0 := yy;
  end;

  lgm := (yp*yp - 3.0)/6.0;

  y := 1.0 / (2.0*a-1.0);
  x := 1.0 / (2.0*b-1.0);
  d := y - x;
  x := 2.0 / (y + x);
  d := d*(lgm + 5.0/6.0 - 2.0/(3.0*x));
  d := 2.0*(d + yp*sqrt(x+lgm)/x);
  if d < ln_MinExt then begin
    {underflow}
    x := 0.0;
    goto done;
  end;

  x  := a/(a+b*exp(d));
  y  := sfc_ibeta(a, b, x);
  yp := (y-y0)/y0;
  if abs(yp) < 0.2 then goto newt;

  {resort to interval halving if not close enough}
ihalve:

  dir := 0;
  di  := 0.5;

  for i:=0 to 400 do begin
    if i<>0 then begin
      x := x0 + di*(x1-x0);
      if x=1.0 then x := 1.0-eps_x;
      if x=0.0 then begin
        di := 0.5;
        x  := x0 + di * (x1 - x0);
        if x=0.0 then goto done; {underflow}
      end;
      y := sfc_ibeta(a, b, x);
      yp := (x1 - x0) / (x1 + x0);
      if abs(yp) < dithresh then goto newt;
      yp := (y-y0)/y0;
      if abs(yp) < dithresh then goto newt;
    end;
    if y<y0 then begin
      x0 := x;
      yl := y;
      if dir < 0 then begin
        dir := 0;
        di  := 0.5;
      end
      else if dir > 3 then di := 1.0 - sqr(1.0-di)
      else if dir > 1 then di := 0.5*di + 0.5
      else di := (y0-y)/(yh-yl);

      inc(dir);

      if x0 > 0.75 then begin
        if rflg = 1 then begin
          rflg := 0;
          a  := aa;
          b  := bb;
          y0 := yy;
        end
        else begin
          rflg := 1;
          a  := bb;
          b  := aa;
          y0 := 1.0 - yy;
        end;
        x  := 1.0 - x;
        y  := sfc_ibeta(a, b, x);
        x0 := 0.0;
        yl := 0.0;
        x1 := 1.0;
        yh := 1.0;
        goto ihalve
      end
    end
    else begin
      x1 := x;
      if (rflg=1) and (x1<eps_x) then begin
        x := 0.0;
        goto done
      end;
      yh := y;
      if dir>0 then begin
        dir := 0;
        di  := 0.5
      end
      else if dir < -3 then di := sqr(di)
      else if dir < -1 then di := 0.5*di
      else di := (y-y0)/(yh - yl);
      dec(dir);
    end;
  end; {for i}

  if RTE_NoConvergence>0 then RunError(byte(RTE_NoConvergence));

  if x0 >= 1.0 then begin
    x := 1.0 - eps_x;
    goto done;
  end;

  if x <= 0.0 then begin
    x := 0.0;
    goto done;
  end;

newt:

  if nflg = 1 then goto done;

  nflg := 1;
  lgm  := -sfc_lnbeta(a, b);
  for i:=0 to 14 do begin
    {compute the function at this point}
    if i<>0 then y := sfc_ibeta(a, b, x);
    if y < yl then begin
      x := x0;
      y := yl;
    end
    else if y > yh then begin
      x := x1;
      y := yh;
    end
    else if y < y0 then begin
      x0 := x;
      yl := y;
    end
    else begin
      x1 := x;
      yh := y;
    end;

    if (x=1.0) or (x=0.0) then goto breaknewt;

    {compute the derivative of the function at this point}
    d := (a-1.0)*ln(x) + (b-1.0)*ln1p(-x) + lgm;
    if d < ln_MinExt then goto done;
    if d > ln_MaxExt then goto breaknewt;
    d := exp(d);

    {compute the step to the next approximation of x}
    d  := (y-y0)/d;
    xt := x - d;
    if xt <= x0 then begin
      y  := (x-x0)/(x1-x0);
      xt := x0 + 0.5*y*(x-x0);
      if xt <= 0.0 then goto breaknewt;
    end;
    if xt >= x1 then begin
      y  := (x1-x)/(x1-x0);
      xt := x1 - 0.5*y*(x1-x);
      if xt >= 1.0 then goto breaknewt;
    end;
    x := xt;
    if abs(d/x) < 128.0*eps_x then goto done
  end;

breaknewt:
  {Newton steps did not converge}
  dithresh := 256.0*eps_x;
  goto ihalve;

done:
  if rflg=1 then begin
    if x <= eps_x then x := 1.0-eps_x
    else x := 1.0-x;
  end;

  sfc_beta_inv := x;
end;


{---------------------------------------------------------------------------}
function sfc_binomial_cdf(p: extended; n, k: longint): extended;
  {-Return the cumulative binomial distribution function with number}
  { of trials n >= 0 and success probability 0 <= p <= 1}
begin
  if (p < 0.0) or (p > 1.0) or (n < 0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_binomial_cdf := NaN_x;
    exit;
  end;
  if k<0 then sfc_binomial_cdf := 0.0
  else if (k>=n) or (p=0.0) then sfc_binomial_cdf := 1.0
  else sfc_binomial_cdf := sfc_ibeta(n-k,k+1,1.0-p)
end;


{---------------------------------------------------------------------------}
function sfc_binomial_pmf(p: extended; n, k: longint): extended;
  {-Return the binomial distribution probability mass function with number}
  { of trials n >= 0 and success probability 0 <= p <= 1}
begin
  if (p < 0.0) or (p > 1.0) or (n < 0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_binomial_pmf := NaN_x;
    exit;
  end;
  if (k<0) or (k>n) then sfc_binomial_pmf := 0.0
  else if n=0 then sfc_binomial_pmf := 1.0
  else if k=0 then sfc_binomial_pmf := power(1.0-p, n)
  else if k=n then sfc_binomial_pmf := power(p, n)
  else begin
    {$ifdef Ver50}
      p := sfc_beta_pdf(k+1,n-k+1,p);
      sfc_binomial_pmf := p/(n+1);
    {$else}
      sfc_binomial_pmf := sfc_beta_pdf(k+1,n-k+1,p)/(n+1);
    {$endif}
  end;
end;


{---------------------------------------------------------------------------}
function sfc_cauchy_pdf(a, b, x: extended): extended;
  {-Return the Cauchy probability density function with location a }
  { and scale b > 0, 1/(Pi*b*(1+((x-a)/b)^2))}
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_cauchy_pdf := NaN_x;
    exit;
  end;
  x := sqr((x-a)/b);
  sfc_cauchy_pdf := 1.0/(Pi*b*(1.0+x));
end;


{---------------------------------------------------------------------------}
function sfc_cauchy_cdf(a, b, x: extended): extended;
  {-Return the cumulative Cauchy distribution function with location a}
  { and scale b > 0, = 1/2 + arctan((x-a)/b)/Pi}
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_cauchy_cdf := NaN_x;
    exit;
  end;
  if x=a then sfc_cauchy_cdf := 0.5
  else begin
    {Definition is inaccurate for (x-a)/b < -1, use arctan identity}
    x := x-a;
    if x <= -b then sfc_cauchy_cdf := arctan2(b,-x)/Pi
    else sfc_cauchy_cdf := 0.5 + arctan2(x,b)/Pi;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_cauchy_inv(a, b, y: extended): extended;
  {-Return the functional inverse of Cauchy distribution function}
  { with location a and scale b > 0}
begin
  if (b <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_cauchy_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_cauchy_inv := PosInf_x
  else if y=0.0 then sfc_cauchy_inv := NegInf_x
  else if y=0.5 then sfc_cauchy_inv := a
  else begin
    if y>0.5 then y := y - 1.0;
    sfc_cauchy_inv := a - b/tan(y*Pi);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_chi2_pdf(nu: longint; x: extended): extended;
  {-Return the probability density function of the chi-square distribution, nu>0}
var
  z: extended;
const
  xmin  = 1e-20;  {DAMath: 2e-17}
  numax = 448;    {DAmath: 39}
begin
  if (nu<1) or (x<0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_chi2_pdf := NaN_x;
    exit;
  end;
  if x<=xmin then begin
    if nu>=numax then sfc_chi2_pdf := 0.0
    else begin
      {pdf = (x/2)^(nu/2-1)/(2*GAMMA(nu/2)); with z = sqrt(0.5*x)}
      {pdf = z^(nu-2)/(2*Gamma(nu/2)),  z = sqrt(2*x)/2          }
      z := 0.5*sqrt(2.0*x);
      z := 0.5*power(z,nu-2);
      sfc_chi2_pdf := z/sfc_gamma(0.5*nu);
    end;
  end
  else sfc_chi2_pdf := sfc_igprefix(0.5*nu,0.5*x)/x;
end;


{---------------------------------------------------------------------------}
function sfc_chi2_cdf(nu: longint; x: extended): extended;
  {-Return the cumulative chi-square distribution with nu>0 degrees of freedom, x >= 0}
begin
  if (nu<1) or (x<0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_chi2_cdf := NaN_x;
    exit;
  end;
  sfc_chi2_cdf := sfc_igammap(0.5*nu, 0.5*x);
end;


{---------------------------------------------------------------------------}
function sfc_chi2_inv(nu: longint; p: extended): extended;
  {-Return the functional inverse of the chi-square distribution, nu>0, 0 <= p < 1}
begin
  if (nu<1) or (p<0.0) or (p>1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_chi2_inv := NaN_x;
    exit;
  end;
  {will return Inf for p=1}
  sfc_chi2_inv := 2.0*sfc_igammap_inv(0.5*nu, p);
end;


{---------------------------------------------------------------------------}
function sfc_evt1_pdf(a, b, x: extended): extended;
  {-Return the probability density function of the Extreme Value Type I distribution}
  { with location a and scale b > 0, result = exp(-(x-a)/b)/b * exp(-exp(-(x-a)/b)) }
var
  t,z: extended;
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_evt1_pdf := NaN_x;
    exit;
  end;
  z := (a-x)/b;
  if z >= ln_MaxExt then sfc_evt1_pdf := 0.0
  else begin
    t := exp(z);
    z := exp(-t);
    sfc_evt1_pdf := t*z/b;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_evt1_cdf(a, b, x: extended): extended;
  {-Return the cumulative Extreme Value Type I distribution function}
  { with location a and scale b > 0;  result = exp(-exp(-(x-a)/b)). }
var
  t,z: extended;
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_evt1_cdf := NaN_x;
    exit;
  end;
  z := (a-x)/b;
  if z >= ln_MaxExt then sfc_evt1_cdf := 0.5
  else begin
    t := exp(z);
    sfc_evt1_cdf := exp(-t);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_evt1_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the Extreme Value Type I distribution}
  { function with location a and scale b > 0;  result = a - b*ln(ln(-y)). }
var
  t: extended;
begin
  if (b <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_evt1_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_evt1_inv := PosInf_x
  else if y=0.0 then sfc_evt1_inv := NegInf_x
  else begin
    t := ln(-ln(y));
    sfc_evt1_inv := a - b*t;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_exp_pdf(a, alpha, x: extended): extended;
  {-Return the exponential probability density function with location a }
  { and rate alpha > 0, = alpha*exp(-alpha*(x-a)) if x >= a, 0 if x < a.}
begin
  if alpha <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_exp_pdf := NaN_x;
    exit;
  end;
  if x < a then sfc_exp_pdf := 0.0
  else sfc_exp_pdf := alpha*exp(-alpha*(x-a));
end;


{---------------------------------------------------------------------------}
function sfc_exp_cdf(a, alpha, x: extended): extended;
  {-Return the cumulative exponential distribution function with location a}
  { and rate alpha > 0, = 1 - exp(-alpha*(x-a)) if x >= a, 0 if x < a.}
begin
  if alpha <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_exp_cdf := NaN_x;
    exit;
  end;
  if x<=a then sfc_exp_cdf := 0.0
  else sfc_exp_cdf := -expm1(-alpha*(x-a));
end;


{---------------------------------------------------------------------------}
function sfc_exp_inv(a, alpha, y: extended): extended;
  {-Return the functional inverse of exponential distribution function with}
  { location a and rate alpha > 0}
begin
  if (alpha <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_exp_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_exp_inv := PosInf_x
  else if y=0.0 then sfc_exp_inv := NegInf_x
  else sfc_exp_inv := a - ln1p(-y)/alpha;
end;


{---------------------------------------------------------------------------}
function sfc_f_pdf(nu1, nu2: longint; x: extended): extended;
  {-Return the probability density function of the F distribution; x >= 0, nu1, nu2 > 0}
var
  a,b,w,z: extended;
begin
  if (x<0.0) or (nu1<1) or (nu2<1) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_f_pdf := NaN_x;
    exit;
  end;
  a := 0.5*nu1;
  b := 0.5*nu2;
  w := a/b;
  z := (a+b)*ln1p(w*x);
  z := z + sfc_lnbeta(a,b);
  w := a*ln(w) + (a-1.0)*ln(x);
  sfc_f_pdf := exp(w-z);
end;


{---------------------------------------------------------------------------}
function sfc_f_cdf(nu1, nu2: longint; x: extended): extended;
  {-Return the cumulative F distribution function; x >= 0, nu1, nu2 > 0}
var
  w,a,b: extended;
begin
  if (x<0.0) or (nu1<1) or (nu2<1) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_f_cdf := NaN_x;
    exit;
  end;
  a := 0.5*nu1;
  b := 0.5*nu2;
  w := a*x;
  w := w/(b+w);
  sfc_f_cdf := sfc_ibeta(a,b,w);
  (*
  if w < b then begin
    w := w/(b+w);
    sfc_f_cdf := sfc_ibeta(a,b,w);
  end
  else begin
    w := b/(b+w);
    sfc_f_cdf := 1.0 - sfc_ibeta(b,a,w);
  end;
  *)
end;


{---------------------------------------------------------------------------}
function sfc_f_inv(nu1, nu2: longint; y: extended): extended;
  {-Return the functional inverse of the F distribution, nu1, nu2 > 0, 0 <= y <= 1}
var
  w,a,b: extended;
begin
  if (y<0.0) or (y>1.0) or (nu1<1) or (nu2<1) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_f_inv := NaN_x;
    exit;
  end;
  a := 0.5*nu1;
  b := 0.5*nu2;
  {Set w := 0.5 to get a slightly faster version, but this }
  {is significant only if sfc_beta_inv will be much faster.}
  w := sfc_ibeta(a, b, 0.5);
  if (w > y) or (y>0.999) then begin
    w := sfc_beta_inv(a, b, y);
    sfc_f_inv := (b*w)/(a*(1.0-w));
  end
  else begin
    w := sfc_beta_inv(b, a, 1.0-y);
    sfc_f_inv := (b*(1.0-w))/(a*w);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_gamma_pdf(a, b, x: extended): extended;
  {-Return the probability density function of a gamma distribution with shape}
  { a>0, scale b>0: sfc_gamma_pdf = x^(a-1)*exp(-x/b)/gamma(a)/b^a, x>0}
begin
  if (a <= 0.0) or (b <= 0.0) or (x < 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_gamma_pdf := NaN_x;
    exit;
  end;
  if x=0.0 then begin
    if a=1.0 then sfc_gamma_pdf := 1.0/b
    else if a>1.0 then sfc_gamma_pdf := 0
    else sfc_gamma_pdf := PosInf_x
  end
  else sfc_gamma_pdf := sfc_igprefix(a,x/b)/x;
end;


{---------------------------------------------------------------------------}
function sfc_gamma_cdf(a, b, x: extended): extended;
  {-Return the cumulative gamma distribution function, shape a>0, scale b>0}
begin
  if (a <= 0.0) or (b <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_gamma_cdf := NaN_x;
    exit;
  end;
  if x<0.0 then sfc_gamma_cdf := 0.0
  else sfc_gamma_cdf := sfc_igammap(a, x/b);
end;


{---------------------------------------------------------------------------}
function sfc_gamma_inv(a, b, p: extended): extended;
  {-Return the functional inverse of the gamma distribution function, shape a>0,}
  { scale b>0, 0 <= p <= 1, i.e. finds x such that gamma_cdf(a, b, x) = p}
begin
  sfc_gamma_inv := sfc_igammap_inv(a, p)*b;
end;


{---------------------------------------------------------------------------}
function bin_raw(p,q: extended; n, k: longint): extended;
  {-Return raw binomial(n,k)*p^k*q^(n-k) without arg check}
var
  lc,lf: extended;

  function bd0(x,np: extended): extended;
    {-Return "deviance part", see Loader article}
  var
    ej, s, s1, v: extended;
    j: integer;
  begin
    if abs(x-np) < 0.1*(x+np) then begin
      v  := (x-np)/(x+np);
      s  := (x-np)*v;
      ej := 2.0*x*v;
      v  := v*v;
      j  := 3;
      repeat
        ej := ej*v;
        s1 := s + ej/j;
        if s=s1 then begin
          bd0 := s1;
          exit;
        end;
        s := s1;
        inc(j,2);
      until j<0;
    end;
    bd0 := x*ln(x/np)+np-x;
  end;

  function stirlerr(k: longint): extended;
    {-sfc_lngcorr for integer, avaoid lngamma for small k}
  const
    sterr: array[0..7] of extended = (
             0,
             0.81061466795327258219670263595e-1,
             0.41340695955409294093822081405e-1,
             0.27677925684998339148789292755e-1,
             0.20790672103765093111522771755e-1,
             0.16644691189821192163194865355e-1,
             0.13876128823070747998745727025e-1,
             0.11896709945891770095055724095e-1);
  begin
    if k>=8 then stirlerr := sfc_lngcorr(k)
    else stirlerr := sterr[k];
  end;

begin
  {Raw binomial pmf used in hypergeometric pmf based on R function dbinom.c and}
  {Catherine Loader (2000). Fast and Accurate Computation of Binomial Probabilities}
  {http://projects.scipy.org/scipy/raw-attachment/ticket/620/loader2000Fast.pdf}
  if (k<0) or (k>n) then bin_raw := 0.0
  else if n=0 then bin_raw:= 1.0
  else if k=0 then bin_raw:= power(q, n)
  else if k=n then bin_raw:= power(p, n)
  else begin
    {$ifdef VER50}
      lc := stirlerr(n);
      lf := stirlerr(k);    lc := lc-lf;
      lf := stirlerr(n-k);  lc := lc -lf;
      lf := bd0(k,n*p);     lc := lc -lf;
      lf := bd0(n-k,n*q);   lc := lc -lf;
    {$else}
      lc := stirlerr(n) - stirlerr(k) - stirlerr(n-k) - bd0(k,n*p) - bd0(n-k,n*q);
    {$endif}
    lf := ln(TwoPi) + ln(k) + ln1p(-k/n);
    bin_raw := exp(lc - 0.5*lf);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_hypergeo_pmf(n1,n2,n,k: longint): extended;
  {-Return the hypergeometric distribution probability mass function; n,n1,n2 >= 0, n <= n1+n2;}
  { i.e. the probability that among n randomly chosen samples from a container}
  { with n1 type1 objects and n2 type2 objects are exactly k type1 objects:}
var
  b1,b2,b3,p,q: extended;
begin
  {Result = binomial(n1,k)*binomial(n2,n-k)/binomial(n1+n2,n)}
  if (n1<0) or (n2<0) or (n<0) or (n>n1+n2) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_hypergeo_pmf := NaN_x;
    exit;
  end;
  sfc_hypergeo_pmf := 0.0;
  if (k<0) or (k>n1) or (k>n) then exit;
  if (n2<n) and (k<n-n2) then exit;
  if n1+n2 = 0 then begin
    {here n=n1=n2=0; result is 1 if k=1 and 0 otherwise}
    if k=0 then sfc_hypergeo_pmf := 1.0;
  end
  else begin
    p  := n/(n1+n2);
    q  := (n1+n2-n)/(n1+n2);
    {R trick: use binomial_pmf with p=n/(n1+n2) instead of binomial()}
    b1 := bin_raw(p,q,n1,k);
    b2 := bin_raw(p,q,n2,n-k);
    b3 := bin_raw(p,q,n1+n2,n);
    sfc_hypergeo_pmf := b1*(b2/b3);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_hypergeo_cdf(n1,n2,n,k: longint): extended;
  {-Return the cumulative hypergeometric distribution function; n,n1,n2 >= 0, n <= n1+n2}
var
  p,t,r,eps: extended;
  i: longint;
begin
  if (n1<0) or (n2<0) or (n<0) or (n>n1+n2) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_hypergeo_cdf := NaN_x;
    exit;
  end;
  if (k >= n1) or (k >= n) then sfc_hypergeo_cdf := 1.0
  else if k<0 then sfc_hypergeo_cdf := 0.0
  else begin
    {Summation of probability mass function using recurrence relations}
    eps := 0.5*eps_x;
    if k >= (n1/(n1+n2))*n then begin
      {use upper tail with ascending recurrence:}
      {f(k+1) = ((n1-k)*(n-k))/((k+1)*(n2-n+k+1)) * f(k)}
      i := succ(k);
      t := sfc_hypergeo_pmf(n1,n2,n,i);
      p := t;
      while (i<n) and (t > eps*p) do begin
        r := ((n1-i)/(i+1)) * ((n-i)/(n2+i+1-n));
        t := t*r;
        p := p+t;
        inc(i);
      end;
      sfc_hypergeo_cdf := 1.0-p;
    end
    else begin
      {use lower tail with descending recurrence:}
      {f(k-1) = (k*(n2-n+k))/((n2-n+k)*(n-k+1)) * f(k)}
      i := k;
      t := sfc_hypergeo_pmf(n1,n2,n,i);
      p := t;
      while (i>0) and (t > eps*p) do begin
        r := (i/(n1-i+1)) * ((n2+i-n)/(n-i+1));
        t := t*r;
        p := p+t;
        dec(i);
      end;
      sfc_hypergeo_cdf := p;
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_kumaraswamy_pdf(a, b, x: extended): extended;
  {-Return the Kumaraswamy probability density function with shape}
  { parameters a,b>0, 0<=x<=1; result = a*b*x^(a-1)*(1-x^a)^(b-1) }
var
  s,t: extended;
begin
  if (a <= 0.0) or (b <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_kumaraswamy_pdf := NaN_x;
    exit;
  end;
  if (x<0.0) or (x>1.0) then sfc_kumaraswamy_pdf := 0.0
  else begin
    s := powm1(x,a);
    t := power(-s, b-1.0);
    s := power(x,a-1.0);
    sfc_kumaraswamy_pdf := a*b*s*t;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_kumaraswamy_cdf(a, b, x: extended): extended;
  {-Return the cumulative Kumaraswamy distribution function with}
  { shape parameters a,b > 0, 0 <= x <= 1; result = 1-(1-x^a)^b}
var
  t: extended;
begin
  if (a <= 0.0) or (b <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_kumaraswamy_cdf := NaN_x;
    exit;
  end;
  if x<=0.0 then sfc_kumaraswamy_cdf := 0.0
  else if x>=1.0 then sfc_kumaraswamy_cdf := 1.0
  else begin
    t := powm1(x,a);
    sfc_kumaraswamy_cdf := -powm1(-t,b);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_kumaraswamy_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the Kumaraswamy distribution}
  { with shape parameters a,b > 0; result = [1-(1-y)^(1/b)]^(1/a)}
var
  t: extended;
begin
  if (a <= 0.0) or (b <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_kumaraswamy_inv := NaN_x;
    exit;
  end;
  if (y=0.0) or (y=1.0) then sfc_kumaraswamy_inv := y
  else begin
    t := pow1pm1(-y,1.0/b);
    sfc_kumaraswamy_inv := power(-t,1.0/a);
  end;
end;



{---------------------------------------------------------------------------}
function sfc_laplace_pdf(a, b, x: extended): extended;
  {-Return the Laplace probability density function with location a}
  { and scale b > 0, result = exp(-abs(x-a)/b) / (2*b)}
var
  y: extended;
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_laplace_pdf := NaN_x;
    exit;
  end;
  y := abs(x-a)/b;
  sfc_laplace_pdf := 0.5*exp(-y)/b;
end;


{---------------------------------------------------------------------------}
function sfc_laplace_cdf(a, b, x: extended): extended;
  {-Return the cumulative Laplace distribution function with location a and scale b > 0}
var
  y: extended;
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_laplace_cdf := NaN_x;
    exit;
  end;
  y := (x-a)/b;
  if y>0.0 then sfc_laplace_cdf := 0.5 - 0.5*expm1(-y)
  else sfc_laplace_cdf := 0.5*exp(y);
end;


{---------------------------------------------------------------------------}
function sfc_laplace_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the Laplace distribution with location a and scale b > 0}
var
  z: extended;
begin
  if (b <= 0.0) or (y<0.0) or (y>1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_laplace_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_laplace_inv := PosInf_x
  else if y=0.0 then sfc_laplace_inv := NegInf_x
  else begin
    if y < 0.5 then sfc_laplace_inv := a + b*ln(2.0*y)
    else begin
      z := 2.0*(1.0-y);
      sfc_laplace_inv := a - b*ln(z);
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_logistic_pdf(a, b, x: extended): extended;
  {-Return the logistic probability density function with location a}
  { and scale parameter b > 0, exp(-(x-a)/b)/b/(1+exp(-(x-a)/b))^2}
var
  y,z: extended;
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_logistic_pdf := NaN_x;
    exit;
  end;
  z := (x-a)/b;
  if abs(z)>ln_MaxExt then sfc_logistic_pdf := 0.0
  else begin
    z := exp(z);
    y := z*b;
    if (y>1.0) and (z>MaxExtended/y) then sfc_logistic_pdf := 1.0/y
    else sfc_logistic_pdf := z/b/sqr(1.0+z);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_logistic_cdf(a, b, x: extended): extended;
  {-Return the cumulative logistic distribution function with}
  { location a and scale parameter b > 0}
var
  z: extended;
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_logistic_cdf := NaN_x;
    exit;
  end;
  z := (a-x)/b;
  if abs(z)>ln_MaxExt then begin
    if z>0.0 then sfc_logistic_cdf := 1.0
    else sfc_logistic_cdf := 0.0;
  end
  else sfc_logistic_cdf := 1.0/(1.0+exp(z))
end;


{---------------------------------------------------------------------------}
function sfc_logistic_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the logistic distribution}
  { with location a and scale parameter b > 0}
begin
  if (b <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_logistic_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_logistic_inv := PosInf_x
  else if y=0.0 then sfc_logistic_inv := NegInf_x
  else begin
    sfc_logistic_inv := a - b*ln((1.0-y)/y);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_lognormal_pdf(a, b, x: extended): extended;
  {-Return the log-normal probability density function with}
  { location a and scale parameter b > 0, zero for x <= 0.}
var
  z: extended;
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_lognormal_pdf := NaN_x;
    exit;
  end;
  if x <= 0.0 then sfc_lognormal_pdf := 0.0
  else begin
    z := (ln(x) - a)/b;
    x := Sqrt_TwoPi*b*x;
    sfc_lognormal_pdf := expmx2h(z)/x;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_lognormal_cdf(a, b, x: extended): extended;
  {-Return the cumulative log-normal distribution function with}
  { location a and scale parameter b > 0, zero for x <= 0.}
begin
  if b <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_lognormal_cdf := NaN_x;
    exit;
  end;
  if x<=0 then sfc_lognormal_cdf := 0.0
  else begin
    x := (ln(x)-a)/b;
    sfc_lognormal_cdf := sfc_normstd_cdf(x);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_lognormal_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the log-normal distribution}
  { with location a and scale parameter b > 0, 0 < y < 1.}
begin
  if (b <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_lognormal_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_lognormal_inv := PosInf_x
  else if y=0.0 then sfc_lognormal_inv := NegInf_x
  else begin
    y := sfc_normstd_inv(y);
    sfc_lognormal_inv := exp(a + b*y);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_maxwell_pdf(b, x: extended): extended;
  {-Return the Maxwell probability density function with scale b > 0, x >= 0}
var
  y: extended;
const
  sqrt2_pi = 0.797884560802865355879892119869;  {sqrt(2/Pi)}
begin
  if (b <= 0.0) or (x < 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_maxwell_pdf := NaN_x;
    exit;
  end;
  {pdf = sqrt(2/Pi) * x^2/b^3 * exp(-0.5*(x/b)^2)}
  x := x/b;
  if x >= Sqrt_MaxExt then sfc_maxwell_pdf := 0.0
  else begin
    y := exp(-0.5*sqr(x));
    y := y*sqr(x)*sqrt2_pi;
    sfc_maxwell_pdf := y/b;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_maxwell_cdf(b, x: extended): extended;
  {-Return the cumulative Maxwell distribution function with scale b > 0, x >= 0}
var
  y: extended;
begin
  if (b <= 0.0) or (x < 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_maxwell_cdf := NaN_x;
    exit;
  end;
  {cdf = igammap(3/2,(x/b)^2/2}
  y := x/b;
  if y >= Sqrt_MaxExt then sfc_maxwell_cdf := 1.0
  else sfc_maxwell_cdf := sfc_igammap(1.5,0.5*sqr(y));
end;


{---------------------------------------------------------------------------}
function sfc_maxwell_inv(b, y: extended): extended;
  {-Return the functional inverse of the Maxwell distribution with scale b > 0}
var
  z: extended;
begin
  if (b <= 0.0) or (y<0.0) or (y>1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_maxwell_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_maxwell_inv := PosInf_x
  else if y=0.0 then sfc_maxwell_inv := 0.0
  else begin
    {inv = b*sqrt(2*igammap_inv(3/2, y))}
    z := sfc_igammap_inv(1.5, y);
    sfc_maxwell_inv := b*sqrt(2.0*z);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_negbinom_cdf(p,r: extended; k: longint): extended;
  {-Return the cumulative negative binomial distribution function with target}
  { for number of successful trials r > 0 and success probability 0 <= p <= 1}
begin
  if (p < 0.0) or (p > 1.0) or (r <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_negbinom_cdf := NaN_x;
    exit;
  end;
  if k<0 then sfc_negbinom_cdf := 0.0
  else sfc_negbinom_cdf := sfc_ibeta(r,k+1,p);
end;


{---------------------------------------------------------------------------}
function sfc_negbinom_pmf(p,r: extended; k: longint): extended;
  {-Return the negative binomial distribution probability mass function with target}
  { for number of successful trials r > 0 and success probability 0 <= p <= 1}
var
  t: extended;
begin
  {sfc_negbinom_pdf = Gamma(k+r)/(k!*Gamma(r))*p^r*(1-p)^k}
  if (p < 0.0) or (p > 1.0) or (r <= 0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_negbinom_pmf := NaN_x;
    exit;
  end;
  if (k<0) or (p=0.0) then sfc_negbinom_pmf := 0.0
  else begin
    t := sfc_beta_pdf(r,k+1,p);
    sfc_negbinom_pmf := (p/(r+k))*t;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_normal_pdf(mu, sd, x: extended): extended;
  {-Return the normal (Gaussian) probability density function with mean mu}
  { and standard deviation sd>0, exp(-0.5*(x-mu)^2/sd^2) / sqrt(2*Pi*sd^2)}
begin
  if sd <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_normal_pdf := NaN_x;
    exit;
  end;
  sfc_normal_pdf := sfc_erf_z((x-mu)/sd)/sd;
end;


{---------------------------------------------------------------------------}
function sfc_normal_cdf(mu, sd, x: extended): extended;
  {-Return the normal (Gaussian) distribution density function}
  { with mean mu and standard deviation sd > 0}
begin
  if sd <= 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_normal_cdf := NaN_x;
    exit;
  end;
  sfc_normal_cdf := sfc_erf_p((x-mu)/sd);
end;


{---------------------------------------------------------------------------}
function sfc_normal_inv(mu, sd, y: extended): extended;
  {-Return the functional inverse of normal (Gaussian) distribution}
  { with mean mu and standard deviation sd > 0, 0 < y < 1.}
begin
  if (sd <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_normal_inv := NaN_x;
    exit;
  end;
  {$ifdef Ver50}
    y := sfc_normstd_inv(y);
    sfc_normal_inv := sd*y + mu;
  {$else}
    sfc_normal_inv := sd*sfc_normstd_inv(y) + mu;
  {$endif}
end;


{---------------------------------------------------------------------------}
function sfc_normstd_pdf(x: extended): extended;
  {-Return the std. normal probability density function exp(-x^2/2)/sqrt(2*Pi)}
begin
  sfc_normstd_pdf := sfc_erf_z(x);
end;


{---------------------------------------------------------------------------}
function sfc_normstd_cdf(x: extended): extended;
  {-Return the standard normal distribution function}
begin
  sfc_normstd_cdf := sfc_erf_p(x);
end;


{---------------------------------------------------------------------------}
function sfc_normstd_inv(y: extended): extended;
  {-Return the inverse standard normal distribution function, 0 <= y <= 1.}
  { For x=normstd_inv(y) and y from [0,1], normstd_cdf(x) = y}
const
  sqrt2h: THexExtW=($6484,$F9DE,$F333,$B504,$3FFF);  {sqrt(2)}
begin
  if IsNanOrInf(y) then sfc_normstd_inv := y
  else if y <= 0.0 then sfc_normstd_inv := NegInf_x
  else if y >= 1.0 then sfc_normstd_inv := PosInf_x
  else sfc_normstd_inv := -sfc_erfc_inv(2.0*y)*extended(sqrt2h);
end;


{---------------------------------------------------------------------------}
function sfc_pareto_pdf(k, a, x: extended): extended;
  {-Return the Pareto probability density function with minimum value k > 0}
  { and shape a, x >= a > 0, result = (a/x)*(k/x)^a}
var
  y: extended;
begin
  if (k <= 0.0) or (a <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_pareto_pdf := NaN_x;
    exit;
  end;
  if x < k then sfc_pareto_pdf := 0.0
  else begin
    y := power(k/x,a);
    sfc_pareto_pdf := y*a/x;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_pareto_cdf(k, a, x: extended): extended;
  {-Return the cumulative Pareto distribution function minimum value k > 0}
  { and shape a, x >= a > 0, result = 1-(k/x)^a}
begin
  if (k <= 0.0) or (a <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_pareto_cdf := NaN_x;
    exit;
  end;
  if x < k then sfc_pareto_cdf := 0.0
  else begin
    sfc_pareto_cdf := -powm1(k/x,a);
    {sfc_pareto_cdf := 1.0 - power(k/x, a);}
    {sfc_pareto_cdf := -expm1(a*ln(k/x));}
  end;
end;


{---------------------------------------------------------------------------}
function sfc_poisson_pmf(mu: extended; k: longint): extended;
  {-Return the Poisson distribution probability mass function with mean mu >= 0}
var
  t: extended;
begin
  if mu<0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_poisson_pmf := NaN_x;
    exit;
  end;
  if k<0.0 then sfc_poisson_pmf := 0.0
  else if mu=0.0 then begin
    if k=0 then sfc_poisson_pmf := 1.0
    else sfc_poisson_pmf := 0.0;
  end
  else begin
    if mu<eps_x then begin
      t := sfc_igprefix(k,mu);
      sfc_poisson_pmf := t/k
    end
    else begin
      t := sfc_igprefix(k+1,mu);
      sfc_poisson_pmf := t/mu;
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_poisson_cdf(mu: extended; k: longint): extended;
  {-Return the cumulative Poisson distribution function with mean mu >= 0}
begin
  if mu < 0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_poisson_cdf := NaN_x;
    exit;
  end;
  if k<0.0 then sfc_poisson_cdf := 0.0
  else if mu=0.0 then sfc_poisson_cdf := 1.0
  else sfc_poisson_cdf := sfc_igammaq(k+1,mu);
end;


{---------------------------------------------------------------------------}
function sfc_pareto_inv(k, a, y: extended): extended;
  {-Return the functional inverse of the Pareto distribution with minimum}
  { value k > 0 and shape a, x >= a > 0, result = k/(1-x)^(1/a)}
begin
  if (k <= 0.0) or (a <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_pareto_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_pareto_inv := PosInf_x
  else if y=0.0 then sfc_pareto_inv := k
  else begin
    sfc_pareto_inv := k/power(1.0-y,1.0/a);
    {sfc_pareto_inv := k*exp(-ln1p(-y)/a);}
  end;
end;


{---------------------------------------------------------------------------}
function sfc_rayleigh_pdf(b, x: extended): extended;
  {-Return the Rayleigh probability density function with}
  { scale b > 0, x >= 0; result = x*exp(-0.5*(x/b)^2)/b^2}
var
  y: extended;
begin
  if (b <= 0.0) or (x < 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_rayleigh_pdf := NaN_x;
    exit;
  end;
  x := x/b;
  if x >= Sqrt_MaxExt then sfc_rayleigh_pdf := 0.0
  else begin
    y := exp(-0.5*sqr(x));
    sfc_rayleigh_pdf := y*x/b;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_rayleigh_cdf(b, x: extended): extended;
  {-Return the cumulative Rayleigh distribution function with scale b > 0, x >= 0}
var
  y: extended;
begin
  if (b <= 0.0) or (x < 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_rayleigh_cdf := NaN_x;
    exit;
  end;
  y := x/b;
  if y >= Sqrt_MaxExt then sfc_rayleigh_cdf := 1.0
  else sfc_rayleigh_cdf := -expm1(-0.5*sqr(y));
end;


{---------------------------------------------------------------------------}
function sfc_rayleigh_inv(b, y: extended): extended;
  {-Return the functional inverse of the Rayleigh distribution with scale b > 0}
var
  z: extended;
begin
  if (b <= 0.0) or (y<0.0) or (y>1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_rayleigh_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_rayleigh_inv := PosInf_x
  else if y=0.0 then sfc_rayleigh_inv := 0.0
  else begin
    z := -2.0*ln1p(-y);
    sfc_rayleigh_inv := b*sqrt(z);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_t_pdf(nu: longint; x: extended): extended;
  {-Return the probability density function of Student's t distribution, nu>0}
var
  b,p: extended;
begin
  if nu<1 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_t_pdf := NaN_x;
    exit;
  end;
  b := sfc_lnbeta(0.5, 0.5*nu);
  p := power(1.0+x*x/nu, -0.5*(nu+1));
  sfc_t_pdf := p*exp(-b)/sqrt(nu);
end;


{---------------------------------------------------------------------------}
function sfc_t_cdf(nu: longint; t: extended): extended;
  {-Return the cumulative Student t distribution with nu>0 degrees of freedom}
var
  x,z,f,tz,p,xsqk: extended;
  j: longint;
begin
  {Ref: Cephes [7], function stdtrl in ldouble/stdtrl.c}
  if nu<1 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_t_cdf := NaN_x;
    exit;
  end;
  if t=0.0 then begin
    sfc_t_cdf := 0.5;
    exit;
  end;
  {Note: Cephes uses only the t < 1.6 condition}
  if (t<-1.6) or (nu>20000) then begin
    if abs(t) >= Sqrt_MaxExt then z := 0.0 else z := nu/(nu+t*t);
    f := 0.5*sfc_ibeta(0.5*nu, 0.5, z);
    if t<=0.0 then sfc_t_cdf := f else sfc_t_cdf := 1.0-f;
    exit;
  end;
  x := abs(t);
  z := 1.0+x*x/nu;
  if odd(nu) then begin
    {computation for odd nu}
    xsqk := x/sqrt(nu);
    p := arctan(xsqk);
    if nu>1 then begin
      f := 1.0;
      tz := 1.0;
      j := 3;
      while (j <= nu-2) and (tz/f > eps_x) do begin
        tz := tz*((j-1)/(z*j));
        f  := f+tz;
        inc(j,2);
      end;
      p := p + f*xsqk/z;
    end;
    p := p/Pi_2;
  end
  else begin
    f  := 1.0;
    tz := 1.0;
    j  := 2;
    while (j <= nu-2) and (tz/f > eps_x) do begin
      tz := tz*((j-1)/(z*j));
      f  := f+tz;
      inc(j,2);
    end;
    p := f*x/sqrt(z*nu);
  end;
  if t<0.0 then begin
    {note destruction of relative accuracy}
    p := -p;
  end;
  sfc_t_cdf:= 0.5+0.5*p;
end;


{---------------------------------------------------------------------------}
function sfc_t_inv(nu: longint; p: extended): extended;
  {-Return the functional inverse of Student's t distribution, nu>0, 0 <= p <= 1}
var
  t,rk: extended;
  rf: boolean;
begin
  {Ref: Cephes[7], function stdtril in ldouble/stdtrl.c}
  if (nu<1) or (p<0.0) or (p>1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_t_inv := NaN_x;
    exit;
  end;

  if p=0.5 then begin
    sfc_t_inv := 0.0;
    exit;
  end;

  rk := nu;
  rf := p<0.5;

  if (p>0.25) and (p<0.75) then begin
    t := sfc_beta_inv(0.5, 0.5*rk, abs(1.0 - 2.0*p));
    t := sqrt(rk*t/(1.0 - t));
  end
  else begin
    if p>0.5 then p := 1.0-p;
    t := sfc_beta_inv(0.5*rk, 0.5, 2.0*p);
    if t*MaxExtended < rk then t := MaxExtended
    else t := sqrt(rk/t - rk);
  end;
  if rf then sfc_t_inv := -t
  else sfc_t_inv := t;
end;


{---------------------------------------------------------------------------}
function sfc_triangular_pdf(a, b, c, x: extended): extended;
  {-Return the triangular probability density function with}
  { lower limit a, upper limit b, mode c;  a<b, a <= c <= b}
var
  t: extended;
begin
  if (a >= b) or (c < a) or (c > b) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_triangular_pdf := NaN_x;
    exit;
  end;
  if (x<a) or (x>b) then sfc_triangular_pdf := 0.0
  else begin
    if x=c then sfc_triangular_pdf := 2.0/(b-a)
    else if (x=a) or (x=b) then sfc_triangular_pdf :=0
    else begin
      if x<c then begin
        t := 0.5*(b-a)*(c-a);
        sfc_triangular_pdf := (x-a)/t
      end
      else begin
        t := 0.5*(b-a)*(b-c);
        sfc_triangular_pdf := (b-x)/t
      end;
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_triangular_cdf(a, b, c, x: extended): extended;
  {-Return the cumulative triangular distribution function with}
  { lower limit a, upper limit b, mode c;  a<b, a <= c <= b}
var
  t: extended;
begin
  if (a >= b) or (c < a) or (c > b) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_triangular_cdf := NaN_x;
    exit;
  end;
  if x < a then sfc_triangular_cdf := 0.0
  else if x >= b then sfc_triangular_cdf := 1.0
  else if x=c then sfc_triangular_cdf := (c-a)/(b-a)
  else if x<c then begin
    t := (b-a)*(c-a);
    sfc_triangular_cdf := sqr(x-a)/t;
  end
  else begin
    t := (b-a)*(b-c);
    t := sqr(b-x)/t;
    sfc_triangular_cdf := 1.0 - t;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_triangular_inv(a, b, c, y: extended): extended;
  {-Return the functional inverse of the triangular distribution with}
  { lower limit a, upper limit b, mode c; a<b, a <= c <= b, 0 <= y <= 1}
var
  t: extended;
begin
  if (a >= b) or (c < a) or (c > b) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_triangular_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_triangular_inv := b
  else if y=0.0 then sfc_triangular_inv := a
  else begin
    t := (c-a)/(b-a);
    if y=t then sfc_triangular_inv := c
    else if y<t then begin
      t := (b-a)*(c-a)*y;
      sfc_triangular_inv := sqrt(t) + a;
    end
    else begin
      t := (b-a)*(b-c)*(1.0-y);
      sfc_triangular_inv := b-sqrt(t);
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_uniform_pdf(a, b, x: extended): extended;
  {-Return the uniform probability density function on [a,b], a<b}
begin
  if (a >= b) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_uniform_pdf := NaN_x;
    exit;
  end;
  if (x<a) or (x>b) then sfc_uniform_pdf := 0.0
  else sfc_uniform_pdf := 1.0/(b-a);
end;


{---------------------------------------------------------------------------}
function sfc_uniform_cdf(a, b, x: extended): extended;
  {-Return the cumulative uniform distribution function on [a,b], a<b}
begin
  if (a >= b) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_uniform_cdf := NaN_x;
    exit;
  end;
  if x < a then sfc_uniform_cdf := 0.0
  else if x >= b then sfc_uniform_cdf := 1.0
  else sfc_uniform_cdf := (x-a)/(b-a);
end;


{---------------------------------------------------------------------------}
function sfc_uniform_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the uniform distribution on [a,b], a<b}
begin
  if (a >= b) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_uniform_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_uniform_inv := b
  else if y=0.0 then sfc_uniform_inv := a
  else sfc_uniform_inv := a + y*(b-a);
end;


{---------------------------------------------------------------------------}
function sfc_weibull_pdf(a, b, x: extended): extended;
  {-Return the Weibull probability density function with shape a > 0}
  { and scale b > 0, result = a*x^(a-1)*exp(-(x/b)^a)/ b^a, x > 0}
var
  y: extended;
begin
  if (b <= 0.0) or (a <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_weibull_pdf := NaN_x;
    exit;
  end;
  if x<0 then sfc_weibull_pdf := 0.0
  else begin
    if x=0.0 then begin
      if a=1.0 then sfc_weibull_pdf := 1.0/b
      else if a>1.0 then sfc_weibull_pdf := 0.0
      else sfc_weibull_pdf := PosInf_x;
    end
    else begin
      y := power(x/b,a);
      y := y*exp(-y);
      sfc_weibull_pdf := y*a/x;
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_weibull_cdf(a, b, x: extended): extended;
  {-Return the cumulative Weibull distribution function with}
  { shape parameter a > 0 and scale parameter b > 0}
var
  y: extended;
begin
  if (b <= 0.0) or (a <= 0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_weibull_cdf := NaN_x;
    exit;
  end;
  if x<=0.0 then sfc_weibull_cdf := 0.0
  else begin
    y := power(x/b, a);
    sfc_weibull_cdf := -expm1(-y);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_weibull_inv(a, b, y: extended): extended;
  {-Return the functional inverse of the Weibull distribution}
  { shape parameter a > 0 and scale parameter b > 0}
var
  z: extended;
begin
  if (b <= 0.0) or (a <= 0.0) or (y < 0.0) or (y > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_weibull_inv := NaN_x;
    exit;
  end;
  if y=1.0 then sfc_weibull_inv := PosInf_x
  else if y=0.0 then sfc_weibull_inv := 0
  else begin
    z := -ln1p(-y);
    sfc_weibull_inv := b*power(z,1.0/a);
  end;
end;


end.
